'From Squeak2.7 of 5 January 2000 [latest update: #1762] on 23 April 2000 at 12:42:05 pm'!
Object subclass: #SimpleGeneticAlgorithm
	instanceVariableNames: 'chromosomes fitness mutationRate survivalRate evaluationBlock '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SoftComputing'!
!SimpleGeneticAlgorithm commentStamp: 'drs 4/23/2000 12:40' prior: 0!
This class implements John H. Holland's simple genetic algorithm.
Chromosomes are stored as bit vectors. The entire population of
chromosomes is stored in a Form in the instance variable 'chromosomes'.

The block assigned with 'SimpleGeneticAlgorithm>>evaluationBlock:'
must return a positive fitness value for the chromosome number passed
into the block.  You will get all kinds of ugly runtime errors if the
fitness value is not >= 0.  The rest of the class assumes that higher
fitness values mean that the chromosome is a "better" answer.

Fields can be extracted from a given chromosome, as positive integers
with the 'SimpleGeneticAlgorithm>>fieldAt:width:' message.  This means
that you need to devise a scheme to encode the optimisation parameters
into bit vectors, and concatenate the bit vectors to form a chromosome.

An example of the usage of this class is given below.  This example
will simultaneously find the maximum of two parabolas.  Admittedly,
this isn't a terribly complex example, but it should give an idea of how
this class can be used.

										-drs, 28 Jan. 2000


"
| s xv yv |

s _ SimpleGeneticAlgorithm new chromosomes: 100 chromosomeWidth: 16.

s evaluationBlock:
	[
		:c |
		xv _ (s fieldAt: (0 @ c) width: 8) - 100.
		yv _ (s fieldAt: (8 @ c) width: 8) - 237.
		xv _ 999999 - (xv * xv).
		yv _ 999999 - (yv * yv).
		xv+yv.
	].

s explore.
s evaluatePopulation.
1 to: 100 do:
	[
		:n |
		s chromosomes displayAt: 10 @ 10.
		s nextGeneration.
		xv _ s fieldAt: (0 @ (s bestChromosome)) width: 8.
		yv _ s fieldAt: (8 @ (s bestChromosome)) width: 8.
		n printString displayAt: (10@280).
		xv printString displayAt: (10@300).
		yv printString displayAt: (10@320).
	]
"!


!SimpleGeneticAlgorithm methodsFor: 'as yet unclassified' stamp: 'drs 1/28/2000 19:21'!
bestChromosome
	"Returns the chromosome index of the chromosome with the highest fitness."

	| bestFitness bestIndex |

	bestFitness _ fitness at: 1.
	bestIndex _ 0.
	0 to: ((chromosomes height) - 1) do:
		[
			:c |
			((fitness at: (c + 1)) > bestFitness) ifTrue:
				[
					bestIndex _ c.
					bestFitness _ (fitness at: (c + 1)).
				]
		].

	^bestIndex.! !

!SimpleGeneticAlgorithm methodsFor: 'as yet unclassified' stamp: 'drs 1/28/2000 18:06'!
crossOver: c1 with: c2 at: xPoint

	| tempC |

	tempC _ chromosomes copy: (Rectangle origin: (0 @ c1) extent: (chromosomes width @ 1)).

	tempC copy: (Rectangle origin: (xPoint @ 0) extent: ((chromosomes width - xPoint) @ 1))
			from: (xPoint @ c2) in: chromosomes rule: Form over.

	^tempC
! !

!SimpleGeneticAlgorithm methodsFor: 'as yet unclassified' stamp: 'drs 3/26/2000 22:09'!
evaluateGenerations: genCount
	"Evaluate the specified number of generations."

	self evaluatePopulation.
	1 to: genCount do:
		[
			:g |
			self nextGeneration.
			g printString displayAt: 10 @ 10.
			self chromosomes displayAt: 10 @ 30.
		].! !

!SimpleGeneticAlgorithm methodsFor: 'as yet unclassified' stamp: 'drs 3/28/2000 16:13'!
evaluatePopulation
	"Evaluate each chromosome and update fitness array."

	0 to: ((chromosomes height) - 1) do:
		[
			:y |
			fitness at: (y + 1) put:
				(evaluationBlock value: y).
			"self halt."
		]! !

!SimpleGeneticAlgorithm methodsFor: 'as yet unclassified' stamp: 'drs 1/28/2000 18:35'!
mateSurvivors
	"Combine survivors using single point crossover to produce offspring.
	Also apply mutation to offspring."

	| cCount sCount rCount parentA parentB offspring cLength |

	cCount _ chromosomes height.
	sCount _ (cCount * survivalRate) asInteger.
	rCount _ cCount - sCount.
	cLength _ chromosomes width.

	1 to: rCount do:
		[
			:r |
			parentA _ (0 to: (sCount - 1)) atRandom.
			parentB _ (0 to: (sCount - 1)) atRandom.
			offspring _ self crossOver: parentA with: parentB at: ((0 to: (cLength - 1)) atRandom).

			offspring _ self mutate: offspring.
			chromosomes copy: (Rectangle origin: (0 @ (sCount + r - 1)) extent: (chromosomes width @ 1)) from: (0 @ 0) in: offspring rule: Form over.
		].
! !

!SimpleGeneticAlgorithm methodsFor: 'as yet unclassified' stamp: 'drs 1/28/2000 16:48'!
mutate: c
	"Randomly flip bits based on mutationRate."

	| cLength randNum mutation |

	randNum _ Random new.
	cLength _ chromosomes width.

	mutation _ c copy: (Rectangle origin: 0@0 extent: cLength @ 1).

	0 to: (cLength - 1) do:
		[
			:b |
			(randNum next < mutationRate) ifTrue:
				[
					"self halt."
					mutation pixelValueAt: (b @ 0) put: (1 - (mutation pixelValueAt: (b @ 0))).
				].
		].
	^mutation! !

!SimpleGeneticAlgorithm methodsFor: 'as yet unclassified' stamp: 'drs 3/20/2000 11:52'!
nextGeneration

	| sCount |

	self selectSurvivors.
	self mateSurvivors.
	self evaluatePopulation.
"
	sCount _ ((chromosomes height) * survivalRate) asInteger.
	sCount to: ((chromosomes height) - 1) do:
		[
			:y |
			fitness at: (y + 1) put:
				(evaluationBlock value: y).
		].
"
	^self.! !


!SimpleGeneticAlgorithm methodsFor: 'initialization' stamp: 'drs 1/28/2000 19:19'!
chromosomes: c chromosomeWidth: w
	"Initialize a new instance of SimpleGeneticAlgorithm"

	| mutation |

	mutationRate _ 0.5.
	survivalRate _ 0.5.
	chromosomes _ Form extent: (w @ c).
	1 to: c do:
		[
			:n |
			mutation _ (chromosomes copy: (Rectangle origin: (0 @ n) extent: ((chromosomes width) @ 1))).
			mutation _ self mutate: mutation.
			chromosomes copy: (Rectangle origin: (0 @ n) extent: (chromosomes width @ 1)) from: (0 @ 0) in: mutation rule: Form over.
		].
	mutationRate _ 0.001.



	fitness _ Array new: c.
	1 to: c do:
		[
			:n |
			fitness at: n put: 0.
		].

	^self.
! !

!SimpleGeneticAlgorithm methodsFor: 'initialization' stamp: 'drs 1/28/2000 16:43'!
chromosomes: c chromosomeWidth: w evaluationBlock: eb
	"Initialize a new instance of SimpleGeneticAlgorithm"

	| mutation |

	mutationRate _ 0.5.
	survivalRate _ 0.5.
	chromosomes _ Form extent: (w @ c).
	1 to: c do:
		[
			:n |
			mutation _ (chromosomes copy: (Rectangle origin: (0 @ n) extent: ((chromosomes width) @ 1))).
			mutation _ self mutate: mutation.
			chromosomes copy: (Rectangle origin: (0 @ n) extent: (chromosomes width @ 1)) from: (0 @ 0) in: mutation rule: Form over.
		].
	mutationRate _ 0.001.



	fitness _ Array new: c.
	1 to: c do:
		[
			:n |
			fitness at: n put: 0.
		].

	^self.
! !


!SimpleGeneticAlgorithm methodsFor: 'accessing' stamp: 'drs 1/28/2000 16:29'!
chromosomes
	^chromosomes! !

!SimpleGeneticAlgorithm methodsFor: 'accessing' stamp: 'drs 1/28/2000 15:56'!
evaluationBlock
	^evaluationBlock! !

!SimpleGeneticAlgorithm methodsFor: 'accessing' stamp: 'drs 1/28/2000 15:56'!
evaluationBlock: eb
	evaluationBlock _ eb.! !

!SimpleGeneticAlgorithm methodsFor: 'accessing' stamp: 'drs 2/23/2000 13:36'!
fastFieldAt: pt width: w
	| val pixPerWord blt |

	(w < 30)
		ifTrue:
			[
				pixPerWord _ 32.
				blt _ BitBlt destForm: (Form extent: pixPerWord@1 depth: 1)
				 	sourceForm: chromosomes
					halftoneForm: nil
					combinationRule: Form over
					destOrigin: (pixPerWord - w)@0
					sourceOrigin: pt
					extent: w@1
					clipRect: (0@0 extent: pixPerWord@1).
				blt destForm bits at: 1 put: 0.
				blt copyBits.
				val _ blt destForm bits at: 1.
			]
		ifFalse:
			[
				val _ 0.
				1 to: w do:
					[
						:n |
						val _ (val bitShift: 1) + (chromosomes pixelValueAt: (pt + ((n - 1) @ 0))).
					].
			].
	^val.! !

!SimpleGeneticAlgorithm methodsFor: 'accessing' stamp: 'drs 2/23/2000 13:40'!
fieldAt: pt width: w
	| val pixPerWord blt |

	(w < 30)
		ifTrue:
			[
				pixPerWord _ 32.
				blt _ BitBlt destForm: (Form extent: pixPerWord@1 depth: 1)
				 	sourceForm: chromosomes
					halftoneForm: nil
					combinationRule: Form over
					destOrigin: (pixPerWord - w)@0
					sourceOrigin: pt
					extent: w@1
					clipRect: (0@0 extent: pixPerWord@1).
				blt destForm bits at: 1 put: 0.
				blt copyBits.
				val _ blt destForm bits at: 1.
			]
		ifFalse:
			[
				val _ 0.
				1 to: w do:
					[
						:n |
						val _ (val bitShift: 1) + (chromosomes pixelValueAt: (pt + ((n - 1) @ 0))).
					].
			].
	^val.! !

!SimpleGeneticAlgorithm methodsFor: 'accessing' stamp: 'drs 2/23/2000 13:31'!
fieldAt: pt width: w put: val
	| tmpVal |

	tmpVal _ val.

	w to: 1 by: -1 do:
		[
			:n |
			chromosomes pixelValueAt: (pt + ((n - 1) @ 0)) put: (tmpVal bitAnd: 1).
			tmpVal _ tmpVal bitShift: -1.
		].

	^val.! !

!SimpleGeneticAlgorithm methodsFor: 'accessing' stamp: 'drs 1/28/2000 15:54'!
mutationRate
	^mutationRate! !

!SimpleGeneticAlgorithm methodsFor: 'accessing' stamp: 'drs 1/28/2000 15:55'!
mutationRate: m
	mutationRate _ m.! !

!SimpleGeneticAlgorithm methodsFor: 'accessing' stamp: 'drs 3/29/2000 17:34'!
selectSurvivors
	"Select survivors based on survivalRate, using Roulette Wheel algorithm."

	| cCount sCount fitSum survivor sRand randGen sSum |
	cCount _ chromosomes height.
	sCount _ (cCount * survivalRate) asInteger.

	randGen _ Random new.

	fitSum _ 0.0.
	1 to: cCount do:
		[
			:c |
			fitSum _ fitSum + (fitness at: c)
		].

	fitSum printString displayAt: 50 @ 10.
	((fitness at: (self bestChromosome + 1)) printString) displayAt: 10 @ 300.


	self swapChromosome: 0 with: (self bestChromosome).
	1 to: (sCount - 1) do:
		[
			:s |
			sSum _ 0.

			"Only pick from chromosomes that are not already chosen to survive."
			survivor _ s.

			sRand _ 1.0e-20 + ((randGen next) * fitSum).
			[(sSum < sRand) and: [survivor < cCount]] whileTrue:
				[
					sSum _ sSum + (fitness at: (survivor + 1)).
					survivor _ survivor + 1.
				].
			self swapChromosome: (survivor - 1) with: s.

			"Only pick from chromosomes that are not already chosen to survive."
			fitSum _ fitSum - (fitness at: s).
		].! !

!SimpleGeneticAlgorithm methodsFor: 'accessing' stamp: 'drs 1/28/2000 15:54'!
survivalRate
	^survivalRate! !

!SimpleGeneticAlgorithm methodsFor: 'accessing' stamp: 'drs 1/28/2000 15:55'!
survivalRate: s
	survivalRate _ s.! !

!SimpleGeneticAlgorithm methodsFor: 'accessing' stamp: 'drs 3/28/2000 16:56'!
swapChromosome: c1 with: c2
	"Exchange chromosome at row c1 with chromosome at row c2."

	| tempC |

	tempC _ fitness at: (c1 + 1).
	fitness at: (c1 + 1) put: (fitness at: (c2 + 1)).
	fitness at: (c2 + 1) put: tempC.

	tempC _ chromosomes copy: (Rectangle origin: (0 @ c1) extent: (chromosomes width @ 1)).
	chromosomes copy: (Rectangle origin: (0 @ c1) extent: (chromosomes width @ 1)) from: (0 @ c2) in: chromosomes rule: Form over.
	chromosomes copy: (Rectangle origin: (0 @ c2) extent: (chromosomes width @ 1)) from: (0 @ 0) in: tempC rule: Form over.

! !
